@@ -1,8 +1,8 @@
 /*
    OpenSSL based Authenticode signing for PE/MSI/Java CAB files.
 
-	 Copyright (C) 2005-2014 Per Allansson <pallansson@gmail.com>
-
+   Copyright (C) 2005-2015 Per Allansson <pallansson@gmail.com>
+   Copyright (C) 2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -31,8 +31,6 @@
    files in the program, then also delete it here.
 */
 
-static const char *rcsid = "$Id: osslsigncode.c,v 1.7.1 2014/07/11 14:14:14 mfive Exp $";
-
 /*
    Implemented with good help from:
 
@@ -61,22 +59,29 @@ static const char *rcsid = "$Id: osslsigncode.c,v 1.7.1 2014/07/11 14:14:14 mfiv
 */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include "config.h"
 #endif
 
 #ifdef HAVE_WINDOWS_H
+#define NOCRYPT
+#define WIN32_LEAN_AND_MEAN
 #include <windows.h>
+typedef unsigned char u_char;
 #endif
 
 #include <stdio.h>
 #include <stdlib.h>
+#ifndef _WIN32
 #include <unistd.h>
+#endif
 #include <string.h>
 #include <time.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+
+#ifndef _WIN32
 #ifdef HAVE_SYS_MMAN_H
 #include <sys/mman.h>
 #endif
@@ -84,6 +89,7 @@ static const char *rcsid = "$Id: osslsigncode.c,v 1.7.1 2014/07/11 14:14:14 mfiv
 #ifdef HAVE_TERMIOS_H
 #include <termios.h>
 #endif
+#endif
 
 #ifdef WITH_GSF
 #include <gsf/gsf-infile-msole.h>
@@ -109,12 +115,25 @@ static const char *rcsid = "$Id: osslsigncode.c,v 1.7.1 2014/07/11 14:14:14 mfiv
 #endif
 
 #ifdef ENABLE_CURL
+#ifdef __CYGWIN__
+#ifndef SOCKET
+#define SOCKET UINT_PTR
+#endif
+#endif
 #include <curl/curl.h>
 
 #define MAX_TS_SERVERS 256
 
 #endif
 
+#ifndef FALSE
+#define FALSE 0
+#endif
+
+#ifndef TRUE
+#define TRUE 1
+#endif
+
 #if defined (HAVE_TERMIOS_H) || defined (HAVE_GETPASS)
 #define PROVIDE_ASKPASS 1
 #endif
@@ -429,7 +448,7 @@ static SpcSpOpusInfo* createOpus(const char *desc, const char *url)
 	if (desc) {
 		info->programName = SpcString_new();
 		info->programName->type = 1;
-		info->programName->value.ascii = M_ASN1_IA5STRING_new();
+		info->programName->value.ascii = ASN1_IA5STRING_new();
 		ASN1_STRING_set((ASN1_STRING *)info->programName->value.ascii,
 						(const unsigned char*)desc, strlen(desc));
 	}
@@ -437,7 +456,7 @@ static SpcSpOpusInfo* createOpus(const char *desc, const char *url)
 	if (url) {
 		info->moreInfo = SpcLink_new();
 		info->moreInfo->type = 0;
-		info->moreInfo->value.url = M_ASN1_IA5STRING_new();
+		info->moreInfo->value.url = ASN1_IA5STRING_new();
 		ASN1_STRING_set((ASN1_STRING *)info->moreInfo->value.url,
 						(const unsigned char*)url, strlen(url));
 	}
@@ -459,6 +478,37 @@ static void tohex(const unsigned char *v, unsigned char *b, int len)
 	b[i*2] = 0x00;
 }
 
+static int add_unauthenticated_blob(PKCS7 *sig)
+{
+	u_char *p = NULL;
+	int len = 1024+4;
+	char prefix[] = "\x0c\x82\x04\x00---BEGIN_BLOB---";  // Length data for ASN1 attribute plus prefix
+	char postfix[] = "---END_BLOB---";
+
+	PKCS7_SIGNER_INFO *si = sk_PKCS7_SIGNER_INFO_value(sig->d.sign->signer_info, 0);
+
+	p = OPENSSL_malloc(len);
+	memset(p, 0, len);
+	memcpy(p, prefix, sizeof(prefix));
+	memcpy(p+len-sizeof(postfix), postfix, sizeof(postfix));
+
+	ASN1_STRING *astr = ASN1_STRING_new();
+	ASN1_STRING_set(astr, p, len);
+
+	int nid = OBJ_create("1.3.6.1.4.1.42921.1.2.1",
+						 "unauthenticatedData",
+						 "unauthenticatedData");
+
+	PKCS7_add_attribute (si, nid, V_ASN1_SEQUENCE, astr);
+
+	OPENSSL_free(p);
+
+	return 0;
+}
+
+
+static int g_verbose = 0;
+
 #ifdef ENABLE_CURL
 
 static int blob_has_nl = 0;
@@ -471,6 +521,19 @@ static size_t curl_write( void *ptr, size_t sz, size_t nmemb, void *stream)
 	return BIO_write((BIO*)stream, ptr, sz*nmemb);
 }
 
+static void print_timestamp_error(const char *url, long http_code)
+{
+    if (http_code != -1) {
+        fprintf(stderr, "Failed to convert timestamp reply from %s; "
+                "HTTP status %ld\n", url, http_code);
+    } else {
+        fprintf(stderr, "Failed to convert timestamp reply from %s; "
+                "no HTTP status available", url);
+    }
+
+    ERR_print_errors_fp(stderr);
+}
+
 /*
   A timestamp request looks like this:
 
@@ -501,7 +564,7 @@ static size_t curl_write( void *ptr, size_t sz, size_t nmemb, void *stream)
 
 */
 
-static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const EVP_MD *md, int verbose)
+static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const EVP_MD *md, int verbose, int noverifypeer)
 {
 	CURL *curl;
 	struct curl_slist *slist = NULL;
@@ -528,6 +591,9 @@ static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const
 	curl_easy_setopt(curl, CURLOPT_URL, url);
 /*	  curl_easy_setopt(curl, CURLOPT_VERBOSE, 42);	*/
 
+	if (noverifypeer)
+		curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, FALSE);
+
 	if (rfc3161) {
 		slist = curl_slist_append(slist, "Content-Type: application/timestamp-query");
 		slist = curl_slist_append(slist, "Accept: application/timestamp-reply");
@@ -541,19 +607,20 @@ static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const
 
 	if (rfc3161) {
 		unsigned char mdbuf[EVP_MAX_MD_SIZE];
-		EVP_MD_CTX mdctx;
+		EVP_MD_CTX *mdctx;
 
-		EVP_MD_CTX_init(&mdctx);
-		EVP_DigestInit(&mdctx, md);
-		EVP_DigestUpdate(&mdctx, si->enc_digest->data, si->enc_digest->length);
-		EVP_DigestFinal(&mdctx, mdbuf, NULL);
+		mdctx = EVP_MD_CTX_new();
+		EVP_DigestInit(mdctx, md);
+		EVP_DigestUpdate(mdctx, si->enc_digest->data, si->enc_digest->length);
+		EVP_DigestFinal(mdctx, mdbuf, NULL);
+		EVP_MD_CTX_free(mdctx);
 
 		TimeStampReq *req = TimeStampReq_new();
 		ASN1_INTEGER_set(req->version, 1);
 		req->messageImprint->digestAlgorithm->algorithm = OBJ_nid2obj(EVP_MD_nid(md));
 		req->messageImprint->digestAlgorithm->parameters = ASN1_TYPE_new();
 		req->messageImprint->digestAlgorithm->parameters->type = V_ASN1_NULL;
-		M_ASN1_OCTET_STRING_set(req->messageImprint->digest, mdbuf, EVP_MD_size(md));
+		ASN1_OCTET_STRING_set(req->messageImprint->digest, mdbuf, EVP_MD_size(md));
 		req->certReq = (void*)0x1;
 
 		len = i2d_TimeStampReq(req, NULL);
@@ -610,16 +677,25 @@ static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const
 	} else {
 		(void)BIO_flush(bin);
 
+		long http_code = -1;
+		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &http_code);
+
+		/*
+		 * At this point we could also look at the response body (and perhaps
+		 * log it if we fail to decode the response):
+		 *
+		 *     char *resp_body = NULL;
+		 *     long resp_body_len = BIO_get_mem_data(bin, &resp_body);
+		 */
+
 		if (rfc3161) {
 			TimeStampResp *reply;
 			(void)BIO_flush(bin);
 			reply = ASN1_item_d2i_bio(ASN1_ITEM_rptr(TimeStampResp), bin, NULL);
 			BIO_free_all(bin);
 			if (!reply) {
-				if (verbose) {
-					fprintf(stderr, "Failed to convert timestamp reply\n");
-					ERR_print_errors_fp(stderr);
-				}
+				if (verbose)
+					print_timestamp_error(url, http_code);
 				return -1;
 			}
 			if (ASN1_INTEGER_get(reply->status->status) != 0) {
@@ -661,10 +737,8 @@ static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const
 			p7 = d2i_PKCS7_bio(b64_bin, NULL);
 			if (p7 == NULL) {
 				BIO_free_all(b64_bin);
-				if (verbose) {
-					fprintf(stderr, "Failed to convert timestamp reply\n");
-					ERR_print_errors_fp(stderr);
-				}
+				if (verbose)
+					print_timestamp_error(url, http_code);
 				return -1;
 			}
 			BIO_free_all(b64_bin);
@@ -701,21 +775,21 @@ static int add_timestamp(PKCS7 *sig, char *url, char *proxy, int rfc3161, const
 	return (int)c;
 }
 
-static int add_timestamp_authenticode(PKCS7 *sig, char **url, int nurls, char *proxy)
+static int add_timestamp_authenticode(PKCS7 *sig, char **url, int nurls, char *proxy, int noverifypeer)
 {
 	int i;
 	for (i=0; i<nurls; i++) {
-		int res = add_timestamp(sig, url[i], proxy, 0, NULL, nurls == 1);
+		int res = add_timestamp(sig, url[i], proxy, 0, NULL, g_verbose || nurls == 1, noverifypeer);
 		if (!res) return 0;
 	}
 	return -1;
 }
 
-static int add_timestamp_rfc3161(PKCS7 *sig, char **url, int nurls, char *proxy, const EVP_MD *md)
+static int add_timestamp_rfc3161(PKCS7 *sig, char **url, int nurls, char *proxy, const EVP_MD *md, int noverifypeer)
 {
 	int i;
 	for (i=0; i<nurls; i++) {
-		int res = add_timestamp(sig, url[i], proxy, 1, md, nurls == 1);
+		int res = add_timestamp(sig, url[i], proxy, 1, md, g_verbose || nurls == 1, noverifypeer);
 		if (!res) return 0;
 	}
 	return -1;
@@ -740,9 +814,6 @@ static void cleanup_lib_state(void)
     EVP_cleanup();
 	CONF_modules_free();
     CRYPTO_cleanup_all_ex_data();
-#if OPENSSL_VERSION_NUMBER > 0x10000000
-	ERR_remove_thread_state(NULL);
-#endif
     ERR_free_strings();
 }
 
@@ -751,7 +822,8 @@ static void usage(const char *argv0)
 	fprintf(stderr,
 			"Usage: %s\n\n\t[ --version | -v ]\n\n"
 			"\t[ sign ]\n"
-			"\t\t( -certs <certfile> -key <keyfile> | -pkcs12 <pkcs12file> )\n"
+			"\t\t( -certs <certfile> -key <keyfile> | -pkcs12 <pkcs12file> |\n"
+			"\t\t  -pkcs11engine <engine> -pkcs11module <module> -certs <certfile> -key <pkcs11 key id>)\n"
 			"\t\t[ -pass <password> ] "
 #ifdef PROVIDE_ASKPASS
 			"[ -askpass ]"
@@ -762,17 +834,25 @@ static void usage(const char *argv0)
 			"\t\t[ -n <desc> ] [ -i <url> ] [ -jp <level> ] [ -comm ]\n"
 			"\t\t[ -ph ]\n"
 #ifdef ENABLE_CURL
-			"\t\t[ -t <timestampurl> [ -t ... ] [ -p <proxy> ]]\n"
-			"\t\t[ -ts <timestampurl> [ -ts ... ] [ -p <proxy> ]]\n"
+			"\t\t[ -t <timestampurl> [ -t ... ] [ -p <proxy> ] [ -noverifypeer ] ]\n"
+			"\t\t[ -ts <timestampurl> [ -ts ... ] [ -p <proxy> ] [ -noverifypeer ] ]\n"
 #endif
+			"\t\t[ -addUnauthenticatedBlob ]\n\n"
 			"\t\t[ -nest ]\n\n"
+			"\t\t[ -verbose ]\n\n"
 			"\t\tMSI specific:\n"
 			"\t\t[ -add-msi-dse ]\n\n"
 			"\t\t[ -in ] <infile> [-out ] <outfile>\n\n"
-			"\textract-signature [ -in ] <infile> [ -out ] <outfile>\n\n"
+			"\textract-signature [ -pem ] [ -in ] <infile> [ -out ] <outfile>\n\n"
 			"\tremove-signature [ -in ] <infile> [ -out ] <outfile>\n\n"
+			"\tattach-signature [ -sigin ] <sigfile> [ -in ] <infile> [ -out ] <outfile>\n\n"
 			"\tverify [ -in ] <infile>\n"
-			"\t\t[ -require-leaf-hash {md5,sha1,sha2(56),sha384,sha512}:XXXXXXXXXXXX... ]\n"
+			"\t\t[ -require-leaf-hash {md5,sha1,sha2(56),sha384,sha512}:XXXXXXXXXXXX... ]\n\n"
+			"\tadd [-addUnauthenticatedBlob] [ -in ] <infile> [ -out ] <outfile>\n"
+#ifdef ENABLE_CURL
+			"\t\t[ -t <timestampurl> [ -t ... ] [ -p <proxy> ] [ -noverifypeer ] ]\n"
+			"\t\t[ -ts <timestampurl> [ -ts ... ] [ -p <proxy> ] [ -noverifypeer ] ]\n"
+#endif
 			"\n"
 			"",
 			argv0);
@@ -811,6 +891,8 @@ typedef enum {
 	CMD_EXTRACT,
 	CMD_REMOVE,
 	CMD_VERIFY,
+	CMD_ADD,
+	CMD_ATTACH,
 } cmd_type_t;
 
 
@@ -838,23 +920,9 @@ static const unsigned char classid_page_hash[] = {
 static unsigned char *calc_page_hash(char *indata, unsigned int peheader, int pe32plus,
 									 unsigned int sigpos, int phtype, unsigned int *phlen);
 
-DECLARE_STACK_OF(ASN1_OCTET_STRING)
-#ifndef sk_ASN1_OCTET_STRING_new_null
-#define sk_ASN1_OCTET_STRING_new_null() SKM_sk_new_null(ASN1_OCTET_STRING)
-#define sk_ASN1_OCTET_STRING_free(st) SKM_sk_free(ASN1_OCTET_STRING, (st))
-#define sk_ASN1_OCTET_STRING_push(st, val) SKM_sk_push(ASN1_OCTET_STRING, (st), (val))
-#define i2d_ASN1_SET_OF_ASN1_OCTET_STRING(st, pp, i2d_func, ex_tag, ex_class, is_set) \
-	SKM_ASN1_SET_OF_i2d(ASN1_OCTET_STRING, (st), (pp), (i2d_func), (ex_tag), (ex_class), (is_set))
-#endif
+DEFINE_STACK_OF(ASN1_OCTET_STRING)
 
-DECLARE_STACK_OF(SpcAttributeTypeAndOptionalValue)
-#ifndef sk_SpcAttributeTypeAndOptionalValue_new_null
-#define sk_SpcAttributeTypeAndOptionalValue_new_null() SKM_sk_new_null(SpcAttributeTypeAndOptionalValue)
-#define sk_SpcAttributeTypeAndOptionalValue_free(st) SKM_sk_free(SpcAttributeTypeAndOptionalValue, (st))
-#define sk_SpcAttributeTypeAndOptionalValue_push(st, val) SKM_sk_push(SpcAttributeTypeAndOptionalValue, (st), (val))
-#define i2d_SpcAttributeTypeAndOptionalValue(st, pp, i2d_func, ex_tag, ex_class, is_set) \
-	SKM_ASN1_SET_OF_i2d(SpcAttributeTypeAndOptionalValue, (st), (pp), (i2d_func), (ex_tag), (ex_class), (is_set))
-#endif
+DEFINE_STACK_OF(SpcAttributeTypeAndOptionalValue)
 
 static SpcLink *get_page_hash_link(int phtype, char *indata, unsigned int peheader, int pe32plus, unsigned int sigpos)
 {
@@ -865,19 +933,17 @@ static SpcLink *get_page_hash_link(int phtype, char *indata, unsigned int pehead
 		exit(-1);
 	}
 
-	ASN1_OCTET_STRING *ostr = M_ASN1_OCTET_STRING_new();
-	M_ASN1_OCTET_STRING_set(ostr, ph, phlen);
+	ASN1_OCTET_STRING *ostr = ASN1_OCTET_STRING_new();
+	ASN1_OCTET_STRING_set(ostr, ph, phlen);
 	free(ph);
 
 	STACK_OF(ASN1_OCTET_STRING) *oset = sk_ASN1_OCTET_STRING_new_null();
 	sk_ASN1_OCTET_STRING_push(oset, ostr);
 	unsigned char *p, *tmp;
 	unsigned int l;
-	l = i2d_ASN1_SET_OF_ASN1_OCTET_STRING(oset, NULL, i2d_ASN1_OCTET_STRING,
-										  V_ASN1_SET, V_ASN1_UNIVERSAL, IS_SET);
+	l = i2d_ASN1_SET_ANY((ASN1_SEQUENCE_ANY *)oset, NULL);
 	tmp = p = OPENSSL_malloc(l);
-	i2d_ASN1_SET_OF_ASN1_OCTET_STRING(oset, &tmp, i2d_ASN1_OCTET_STRING,
-									  V_ASN1_SET, V_ASN1_UNIVERSAL, IS_SET);
+	i2d_ASN1_SET_ANY((ASN1_SEQUENCE_ANY *)oset, &tmp);
 	ASN1_OCTET_STRING_free(ostr);
 	sk_ASN1_OCTET_STRING_free(oset);
 
@@ -891,17 +957,15 @@ static SpcLink *get_page_hash_link(int phtype, char *indata, unsigned int pehead
 
 	STACK_OF(SpcAttributeTypeAndOptionalValue) *aset = sk_SpcAttributeTypeAndOptionalValue_new_null();
 	sk_SpcAttributeTypeAndOptionalValue_push(aset, aval);
-	l = i2d_SpcAttributeTypeAndOptionalValue(aset, NULL, i2d_SpcAttributeTypeAndOptionalValue,
-											 V_ASN1_SET, V_ASN1_UNIVERSAL, IS_SET);
+	l = i2d_ASN1_SET_ANY((ASN1_SEQUENCE_ANY *)aset, NULL);
 	tmp = p = OPENSSL_malloc(l);
-	l = i2d_SpcAttributeTypeAndOptionalValue(aset, &tmp, i2d_SpcAttributeTypeAndOptionalValue,
-											 V_ASN1_SET, V_ASN1_UNIVERSAL, IS_SET);
+	l = i2d_ASN1_SET_ANY((ASN1_SEQUENCE_ANY *)aset, &tmp);
 	sk_SpcAttributeTypeAndOptionalValue_free(aset);
 	SpcAttributeTypeAndOptionalValue_free(aval);
 
 	SpcSerializedObject *so = SpcSerializedObject_new();
-	M_ASN1_OCTET_STRING_set(so->classId, classid_page_hash, sizeof(classid_page_hash));
-	M_ASN1_OCTET_STRING_set(so->serializedData, p, l);
+	ASN1_OCTET_STRING_set(so->classId, classid_page_hash, sizeof(classid_page_hash));
+	ASN1_OCTET_STRING_set(so->serializedData, p, l);
 	OPENSSL_free(p);
 
 	SpcLink *link = SpcLink_new();
@@ -960,7 +1024,7 @@ static void get_indirect_data_blob(u_char **blob, int *len, const EVP_MD *md, fi
 		ASN1_INTEGER_set(si->d, 0);
 		ASN1_INTEGER_set(si->e, 0);
 		ASN1_INTEGER_set(si->f, 0);
-		M_ASN1_OCTET_STRING_set(si->string, msistr, sizeof(msistr));
+		ASN1_OCTET_STRING_set(si->string, msistr, sizeof(msistr));
 		l = i2d_SpcSipInfo(si, NULL);
 		p = OPENSSL_malloc(l);
 		i2d_SpcSipInfo(si, &p);
@@ -982,7 +1046,7 @@ static void get_indirect_data_blob(u_char **blob, int *len, const EVP_MD *md, fi
 	hashlen = EVP_MD_size(md);
 	hash = OPENSSL_malloc(hashlen);
 	memset(hash, 0, hashlen);
-	M_ASN1_OCTET_STRING_set(idc->messageDigest->digest, hash, hashlen);
+	ASN1_OCTET_STRING_set(idc->messageDigest->digest, hash, hashlen);
 	OPENSSL_free(hash);
 
 	*len  = i2d_SpcIndirectDataContent(idc, NULL);
@@ -995,19 +1059,29 @@ static void get_indirect_data_blob(u_char **blob, int *len, const EVP_MD *md, fi
 static unsigned int calc_pe_checksum(BIO *bio, unsigned int peheader)
 {
 	unsigned int checkSum = 0;
-	unsigned short	val;
+	unsigned short val;
 	unsigned int size = 0;
+	unsigned short *buf;
+	int nread;
 
 	/* recalc checksum. */
+	buf = (unsigned short*)malloc(sizeof(unsigned short)*32768);
+
 	(void)BIO_seek(bio, 0);
-	while (BIO_read(bio, &val, 2) == 2) {
-		if (size == peheader + 88 || size == peheader + 90)
-			val = 0;
-		checkSum += val;
-		checkSum = 0xffff & (checkSum + (checkSum >> 0x10));
-		size += 2;
+	while ((nread = BIO_read(bio, buf, sizeof(unsigned short)*32768)) > 0) {
+		int i;
+		for (i = 0; i < nread / 2; i++) {
+			val = buf[i];
+			if (size == peheader + 88 || size == peheader + 90)
+				val = 0;
+			checkSum += val;
+			checkSum = 0xffff & (checkSum + (checkSum >> 0x10));
+			size += 2;
+		}
 	}
 
+	free(buf);
+
 	checkSum = 0xffff & (checkSum + (checkSum >> 0x10));
 	checkSum += size;
 
@@ -1827,19 +1901,19 @@ static void calc_pe_digest(BIO *bio, const EVP_MD *md, unsigned char *mdbuf,
 						   unsigned int peheader, int pe32plus, unsigned int fileend)
 {
 	static unsigned char bfb[16*1024*1024];
-	EVP_MD_CTX mdctx;
+	EVP_MD_CTX *mdctx;
 
-	EVP_MD_CTX_init(&mdctx);
-	EVP_DigestInit(&mdctx, md);
+	mdctx = EVP_MD_CTX_new();
+	EVP_DigestInit(mdctx, md);
 
 	memset(mdbuf, 0, EVP_MAX_MD_SIZE);
 
 	(void)BIO_seek(bio, 0);
 	BIO_read(bio, bfb, peheader + 88);
-	EVP_DigestUpdate(&mdctx, bfb, peheader + 88);
+	EVP_DigestUpdate(mdctx, bfb, peheader + 88);
 	BIO_read(bio, bfb, 4);
 	BIO_read(bio, bfb, 60+pe32plus*16);
-	EVP_DigestUpdate(&mdctx, bfb, 60+pe32plus*16);
+	EVP_DigestUpdate(mdctx, bfb, 60+pe32plus*16);
 	BIO_read(bio, bfb, 8);
 
 	unsigned int n = peheader + 88 + 4 + 60+pe32plus*16 + 8;
@@ -1850,11 +1924,12 @@ static void calc_pe_digest(BIO *bio, const EVP_MD *md, unsigned char *mdbuf,
 		int l = BIO_read(bio, bfb, want);
 		if (l <= 0)
 			break;
-		EVP_DigestUpdate(&mdctx, bfb, l);
+		EVP_DigestUpdate(mdctx, bfb, l);
 		n += l;
 	}
 
-	EVP_DigestFinal(&mdctx, mdbuf, NULL);
+	EVP_DigestFinal(mdctx, mdbuf, NULL);
+	EVP_MD_CTX_free(mdctx);
 }
 
 
@@ -1923,16 +1998,16 @@ static unsigned char *calc_page_hash(char *indata, unsigned int peheader, int pe
 	int phlen = pphlen * (3 + nsections + sigpos / pagesize);
 	unsigned char *res = malloc(phlen);
 	unsigned char *zeroes = calloc(pagesize, 1);
-	EVP_MD_CTX mdctx;
+	EVP_MD_CTX *mdctx;
 
-	EVP_MD_CTX_init(&mdctx);
-	EVP_DigestInit(&mdctx, md);
-	EVP_DigestUpdate(&mdctx, indata, peheader + 88);
-	EVP_DigestUpdate(&mdctx, indata + peheader + 92, 60 + pe32plus*16);
-	EVP_DigestUpdate(&mdctx, indata + peheader + 160 + pe32plus*16, hdrsize - (peheader + 160 + pe32plus*16));
-	EVP_DigestUpdate(&mdctx, zeroes, pagesize - hdrsize);
+	mdctx = EVP_MD_CTX_new();
+	EVP_DigestInit(mdctx, md);
+	EVP_DigestUpdate(mdctx, indata, peheader + 88);
+	EVP_DigestUpdate(mdctx, indata + peheader + 92, 60 + pe32plus*16);
+	EVP_DigestUpdate(mdctx, indata + peheader + 160 + pe32plus*16, hdrsize - (peheader + 160 + pe32plus*16));
+	EVP_DigestUpdate(mdctx, zeroes, pagesize - hdrsize);
 	memset(res, 0, 4);
-	EVP_DigestFinal(&mdctx, res + 4, NULL);
+	EVP_DigestFinal(mdctx, res + 4, NULL);
 
 	unsigned short sizeofopthdr = GET_UINT16_LE(indata + peheader + 20);
 	char *sections = indata + peheader + 24 + sizeofopthdr;
@@ -1944,18 +2019,19 @@ static unsigned char *calc_page_hash(char *indata, unsigned int peheader, int pe
 		unsigned int l;
 		for (l=0; l < rs; l+=pagesize, pi++) {
 			PUT_UINT32_LE(ro + l, res + pi*pphlen);
-			EVP_DigestInit(&mdctx, md);
+			EVP_DigestInit(mdctx, md);
 			if (rs - l < pagesize) {
-				EVP_DigestUpdate(&mdctx, indata + ro + l, rs - l);
-				EVP_DigestUpdate(&mdctx, zeroes, pagesize - (rs - l));
+				EVP_DigestUpdate(mdctx, indata + ro + l, rs - l);
+				EVP_DigestUpdate(mdctx, zeroes, pagesize - (rs - l));
 			} else {
-				EVP_DigestUpdate(&mdctx, indata + ro + l, pagesize);
+				EVP_DigestUpdate(mdctx, indata + ro + l, pagesize);
 			}
-			EVP_DigestFinal(&mdctx, res + pi*pphlen + 4, NULL);
+			EVP_DigestFinal(mdctx, res + pi*pphlen + 4, NULL);
 		}
 		lastpos = ro + rs;
 		sections += 40;
 	}
+	EVP_MD_CTX_free(mdctx);
 	PUT_UINT32_LE(lastpos, res + pi*pphlen);
 	memset(res + pi*pphlen + 4, 0, EVP_MD_size(md));
 	pi++;
@@ -2049,9 +2125,16 @@ static int verify_pe_pkcs7(PKCS7 *p7, char *indata, unsigned int peheader, int p
 		X509 *cert = sk_X509_value(signers, i);
 		char *subject = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
 		char *issuer = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
-		printf("\tSigner #%d:\n\t\tSubject: %s\n\t\tIssuer : %s\n", i, subject, issuer);
+		BIGNUM *serialbn = ASN1_INTEGER_to_BN(X509_get_serialNumber(cert), NULL);
+		char *serial = BN_bn2hex(serialbn);
+		if (i > 0)
+			printf("\t------------------\n");
+		printf("\tSigner #%d:\n\t\tSubject: %s\n\t\tIssuer : %s\n\t\tSerial : %s\n",
+			   i, subject, issuer, serial);
 		OPENSSL_free(subject);
 		OPENSSL_free(issuer);
+		OPENSSL_free(serial);
+		BN_free(serialbn);
 
 		if (leafhash != NULL && leafok == 0) {
 			leafok = verify_leaf_hash(cert, leafhash) == 0;
@@ -2064,9 +2147,16 @@ static int verify_pe_pkcs7(PKCS7 *p7, char *indata, unsigned int peheader, int p
 		X509 *cert = sk_X509_value(p7->d.sign->cert, i);
 		char *subject = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
 		char *issuer = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
-		printf("\tCert #%d:\n\t\tSubject: %s\n\t\tIssuer : %s\n", i, subject, issuer);
+		BIGNUM *serialbn = ASN1_INTEGER_to_BN(X509_get_serialNumber(cert), NULL);
+		char *serial = BN_bn2hex(serialbn);
+		if (i > 0)
+			printf("\t------------------\n");
+		printf("\tCert #%d:\n\t\tSubject: %s\n\t\tIssuer : %s\n\t\tSerial : %s\n",
+			   i, subject, issuer, serial);
 		OPENSSL_free(subject);
 		OPENSSL_free(issuer);
+		OPENSSL_free(serial);
+		BN_free(serialbn);
 	}
 
 	if (leafhash != NULL) {
@@ -2191,13 +2281,15 @@ static STACK_OF(X509) *PEM_read_certs(BIO *bin, char *certpass)
 
 static off_t get_file_size(const char *infile)
 {
-#ifdef WIN32
+	int ret;
+#ifdef _WIN32
 	struct _stat st;
-	if (_stat(infile, &st))
+	ret = _stat(infile, &st);
 #else
 	struct stat st;
-	if (stat(infile, &st))
+	ret = stat(infile, &st);
 #endif
+	if (ret)
 	{
 		fprintf(stderr, "Failed to open file: %s\n", infile);
 		return 0;
@@ -2215,7 +2307,7 @@ static char* map_file(const char *infile, const off_t size)
 	char *indata = NULL;
 #ifdef WIN32
 	HANDLE fh, fm;
-	fh = CreateFile(infile, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
+	fh = CreateFile(infile, GENERIC_READ, FILE_SHARE_READ , NULL, OPEN_EXISTING, 0, NULL);
 	if (fh == INVALID_HANDLE_VALUE)
 		return NULL;
 	fm = CreateFileMapping(fh, NULL, PAGE_READONLY, 0, 0, NULL);
@@ -2286,20 +2378,24 @@ int main(int argc, char **argv)
 
 	const char *argv0 = argv[0];
 	static char buf[64*1024];
-	char *xcertfile, *certfile, *keyfile, *pvkfile, *pkcs12file, *infile, *outfile, *desc, *url, *indata;
+	char *xcertfile, *certfile, *keyfile, *pvkfile, *pkcs12file, *infile, *outfile, *sigfile, *desc, *url, *indata, *insigdata, *outdataverify;
+	char *p11engine, *p11module;
 	char *pass = NULL, *readpass = NULL;
+	int output_pkcs7 = 0;
 	int askpass = 0;
 	char *leafhash = NULL;
 #ifdef ENABLE_CURL
 	char *turl[MAX_TS_SERVERS], *proxy = NULL, *tsurl[MAX_TS_SERVERS];
+	int noverifypeer = 0;
 #endif
 	int nest = 0;
 	int add_msi_dse = 0;
 	int nturl = 0, ntsurl = 0;
+	int addBlob = 0;
 	u_char *p = NULL;
 	int ret = 0, i, len = 0, jp = -1, pe32plus = 0, comm = 0, pagehash = 0;
 	unsigned int tmp, peheader = 0, padlen = 0;
-	off_t filesize, fileend;
+	off_t filesize, fileend, sigfilesize, sigfileend, outdatasize;
 	file_type_t type;
 	cmd_type_t cmd = CMD_SIGN;
 	char *failarg = NULL;
@@ -2325,7 +2421,7 @@ int main(int argc, char **argv)
 	int len_msiex = 0;
 #endif
 
-	xcertfile = certfile = keyfile = pvkfile = pkcs12file = infile = outfile = desc = url = NULL;
+	xcertfile = certfile = keyfile = pvkfile = pkcs12file = p11module = p11engine = infile = outfile = sigfile = desc = url = NULL;
 	hash = outdata = NULL;
 
 	/* Set up OpenSSL */
@@ -2350,6 +2446,10 @@ int main(int argc, char **argv)
 			cmd = CMD_EXTRACT;
 			argv++;
 			argc--;
+		} else if (!strcmp(argv[1], "attach-signature")) {
+			cmd = CMD_ATTACH;
+			argv++;
+                        argc--;
 		} else if (!strcmp(argv[1], "remove-signature")) {
 			cmd = CMD_REMOVE;
 			argv++;
@@ -2358,6 +2458,10 @@ int main(int argc, char **argv)
 			cmd = CMD_VERIFY;
 			argv++;
 			argc--;
+		} else if (!strcmp(argv[1], "add")) {
+			cmd = CMD_ADD;
+			argv++;
+			argc--;
 		}
 	}
 
@@ -2368,6 +2472,9 @@ int main(int argc, char **argv)
 		} else if (!strcmp(*argv, "-out")) {
 			if (--argc < 1) usage(argv0);
 			outfile = *(++argv);
+		} else if (!strcmp(*argv, "-sigin")) {
+			if (--argc < 1) usage(argv0);
+			sigfile = *(++argv);
 		} else if ((cmd == CMD_SIGN) && (!strcmp(*argv, "-spc") || !strcmp(*argv, "-certs"))) {
 			if (--argc < 1) usage(argv0);
 			certfile = *(++argv);
@@ -2380,6 +2487,14 @@ int main(int argc, char **argv)
 		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-pkcs12")) {
 			if (--argc < 1) usage(argv0);
 			pkcs12file = *(++argv);
+		} else if ((cmd == CMD_EXTRACT) && !strcmp(*argv, "-pem")) {
+			output_pkcs7 = 1;
+		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-pkcs11engine")) {
+			if (--argc < 1) usage(argv0);
+			p11engine = *(++argv);
+		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-pkcs11module")) {
+			if (--argc < 1) usage(argv0);
+			p11module = *(++argv); 
 		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-pass")) {
 			if (askpass || readpass) usage(argv0);
 			if (--argc < 1) usage(argv0);
@@ -2430,14 +2545,20 @@ int main(int argc, char **argv)
 		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-p")) {
 			if (--argc < 1) usage(argv0);
 			proxy = *(++argv);
+		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-noverifypeer")) {
+			noverifypeer = 1;
 #endif
-		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-nest")) {
+		} else if ((cmd == CMD_SIGN || cmd == CMD_ADD) && !strcmp(*argv, "-addUnauthenticatedBlob")) {
+			addBlob = 1;
+		} else if ((cmd == CMD_SIGN || cmd == CMD_ATTACH) && !strcmp(*argv, "-nest")) {
 			nest = 1;
+		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-verbose")) {
+			g_verbose = 1;
 		} else if ((cmd == CMD_SIGN) && !strcmp(*argv, "-add-msi-dse")) {
 			add_msi_dse = 1;
 		} else if ((cmd == CMD_VERIFY) && !strcmp(*argv, "-require-leaf-hash")) {
 			if (--argc < 1) usage(argv0);
-			leafhash = (*++argv); 
+			leafhash = (*++argv);
 		} else if (!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
 			printf(PACKAGE_STRING ", using:\n\t%s\n\t%s\n",
 				   SSLeay_version(SSLEAY_VERSION),
@@ -2498,7 +2619,7 @@ int main(int argc, char **argv)
 
 	if (argc > 0 || (nturl && ntsurl) || !infile ||
 		(cmd != CMD_VERIFY && !outfile) ||
-		(cmd == CMD_SIGN && !((certfile && keyfile) || pkcs12file))) {
+		(cmd == CMD_SIGN && !((certfile && keyfile) || pkcs12file || (p11engine && p11module)))) {
 		if (failarg)
 			fprintf(stderr, "Unknown option: %s\n", failarg);
 		usage(argv0);
@@ -2509,7 +2630,7 @@ int main(int argc, char **argv)
 		int passfd = open(readpass, O_RDONLY);
 		if (passfd < 0)
 			DO_EXIT_1("Failed to open password file: %s\n", readpass);
-		ssize_t passlen = read(passfd, passbuf, sizeof(passbuf)-1);
+		int passlen = read(passfd, passbuf, sizeof(passbuf)-1);
 		close(passfd);
 		if (passlen <= 0)
 			DO_EXIT_1("Failed to read password from file: %s\n", readpass);
@@ -2524,7 +2645,7 @@ int main(int argc, char **argv)
 
 	if (cmd == CMD_SIGN) {
 		/* Read certificate and key */
-		if (keyfile && (btmp = BIO_new_file(keyfile, "rb")) != NULL) {
+		if (keyfile && !p11engine && (btmp = BIO_new_file(keyfile, "rb")) != NULL) {
 			unsigned char magic[4];
 			unsigned char pvkhdr[4] = { 0x1e, 0xf1, 0xb5, 0xb0 };
 			magic[0] = 0x00;
@@ -2545,7 +2666,6 @@ int main(int argc, char **argv)
 				DO_EXIT_1("Failed to parse PKCS#12 file: %s (Wrong password?)\n", pkcs12file);
 			PKCS12_free(p12);
 		} else if (pvkfile != NULL) {
-#if OPENSSL_VERSION_NUMBER > 0x10000000
 			if ((btmp = BIO_new_file(certfile, "rb")) == NULL ||
 				((p7 = d2i_PKCS7_bio(btmp, NULL)) == NULL &&
 				 (certs = PEM_read_certs(btmp, "")) == NULL))
@@ -2557,9 +2677,47 @@ int main(int argc, char **argv)
 				  (pkey = b2i_PVK_bio(btmp, NULL, NULL)) == NULL))
 				DO_EXIT_1("Failed to read PVK file: %s\n", pvkfile);
 			BIO_free(btmp);
-#else
-			DO_EXIT_1("Can not read keys from PVK files, must compile against a newer version of OpenSSL: %s\n", pvkfile);
-#endif
+		} else if (p11engine != NULL && p11module != NULL) {
+			const int CMD_MANDATORY = 0;
+			ENGINE_load_dynamic();
+			ENGINE * dyn = ENGINE_by_id( "dynamic" );
+			if ( ! dyn )
+				DO_EXIT_0( "Failed to load 'dynamic' engine");
+			if ( 1 != ENGINE_ctrl_cmd_string( dyn, "SO_PATH", p11engine, CMD_MANDATORY ) )
+				DO_EXIT_1( "Failed to set dyn SO_PATH to '%s'", p11engine);
+
+			if ( 1 != ENGINE_ctrl_cmd_string( dyn, "ID", "pkcs11", CMD_MANDATORY ) )
+				DO_EXIT_0( "Failed to set dyn ID to 'pkcs11'" );
+
+			if ( 1 != ENGINE_ctrl_cmd( dyn, "LIST_ADD", 1, NULL, NULL, CMD_MANDATORY ) )
+				DO_EXIT_0( "Failed to set dyn LIST_ADD to '1'" );
+
+			if ( 1 != ENGINE_ctrl_cmd( dyn, "LOAD", 1, NULL, NULL, CMD_MANDATORY ) )
+				DO_EXIT_0( "Failed to set dyn LOAD to '1'" );
+
+			ENGINE * pkcs11 = ENGINE_by_id( "pkcs11" );
+			if ( ! pkcs11 )
+				DO_EXIT_0( "Failed to find and load pkcs11 engine" );
+
+			if ( 1 != ENGINE_ctrl_cmd_string( pkcs11, "MODULE_PATH", p11module, CMD_MANDATORY ) )
+				DO_EXIT_1( "Failed to set pkcs11 engine MODULE_PATH to '%s'", p11module );
+		
+			if (pass != NULL) {
+				if ( 1 != ENGINE_ctrl_cmd_string( pkcs11, "PIN", pass, CMD_MANDATORY ) )
+					DO_EXIT_0( "Failed to set pkcs11 PIN" );
+			}
+
+			if ( 1 != ENGINE_init( pkcs11 ) )
+				DO_EXIT_0( "Failed to initialized pkcs11 engine" );
+
+			pkey = ENGINE_load_private_key( pkcs11, keyfile, NULL, NULL );
+			if (pkey == NULL)
+				DO_EXIT_1("Failed to load private key %s", keyfile);
+			if ((btmp = BIO_new_file(certfile, "rb")) == NULL ||
+				((p7 = d2i_PKCS7_bio(btmp, NULL)) == NULL &&
+				 (certs = PEM_read_certs(btmp, "")) == NULL))
+				DO_EXIT_1("Failed to read certificate file: %s\n", certfile);
+			BIO_free(btmp);
 		} else {
 			if ((btmp = BIO_new_file(certfile, "rb")) == NULL ||
 				((p7 = d2i_PKCS7_bio(btmp, NULL)) == NULL &&
@@ -2650,13 +2808,24 @@ int main(int argc, char **argv)
 		ole = gsf_infile_msole_new(src, NULL);
 
 		if (cmd == CMD_EXTRACT) {
-			ret = msi_extract_signature_to_file(ole, outfile);
+			if(output_pkcs7) {
+				sig = msi_extract_signature_to_pkcs7(ole);
+				if (!sig)
+					DO_EXIT_0("Unable to extract existing signature.");
+				outdata = BIO_new_file(outfile, "w+b");
+				if (outdata == NULL)
+					DO_EXIT_1("Unable to open %s\n\n", outfile);
+				ret = !PEM_write_bio_PKCS7(outdata, sig);
+				BIO_free_all(outdata);
+			}
+			else
+				ret = msi_extract_signature_to_file(ole, outfile);
 			goto skip_signing;
 		} else if (cmd == CMD_VERIFY) {
 			ret = msi_verify_file(ole, leafhash);
 			goto skip_signing;
-		} else if (cmd == CMD_SIGN) {
-			if (nest) {
+		} else if (cmd == CMD_SIGN || cmd == CMD_ADD || cmd == CMD_ATTACH) {
+				if (nest || cmd == CMD_ADD) {
 				// Perform a sanity check for the MsiDigitalSignatureEx section.
 				// If the file we're attempting to sign has an MsiDigitalSignatureEx
 				// section, we can't add a nested signature of a different MD type
@@ -2681,6 +2850,9 @@ int main(int argc, char **argv)
 				if (cursig == NULL) {
 					DO_EXIT_0("Unable to extract existing signature in -nest mode");
 				}
+				if (cmd == CMD_ADD) {
+					sig = cursig;
+				}
 			}
 		}
 
@@ -2845,15 +3017,25 @@ int main(int argc, char **argv)
 		if (cmd == CMD_EXTRACT) {
 			/* A lil' bit of ugliness. Reset stream, write signature and skip forward */
 			(void)BIO_reset(outdata);
-			BIO_write(outdata, indata + sigpos, siglen);
+			if(output_pkcs7) {
+				sig = extract_existing_pe_pkcs7(indata, peheader, pe32plus, sigpos ? sigpos : fileend, siglen);
+				if (!sig)
+					DO_EXIT_0("Unable to extract existing signature.");
+				PEM_write_bio_PKCS7(outdata, sig);
+			}
+			else
+				BIO_write(outdata, indata + sigpos, siglen);
 			goto skip_signing;
 		}
 
-		if (cmd == CMD_SIGN && nest) {
+		if ((cmd == CMD_SIGN && nest) || (cmd == CMD_ATTACH && nest) || cmd == CMD_ADD) {
 			cursig = extract_existing_pe_pkcs7(indata, peheader, pe32plus, sigpos ? sigpos : fileend, siglen);
 			if (cursig == NULL) {
 				DO_EXIT_0("Unable to extract existing signature in -nest mode");
 			}
+			if (cmd == CMD_ADD) {
+				sig = cursig;
+			}
 		}
 
 		if (cmd == CMD_VERIFY) {
@@ -2888,6 +3070,43 @@ int main(int argc, char **argv)
 		}
 	}
 
+	if (cmd == CMD_ADD)
+		goto add_only;
+
+	if(cmd == CMD_ATTACH)
+	{
+		const char pemhdr[] = "-----BEGIN PKCS7-----";
+		sigfilesize = get_file_size(sigfile);
+		if(!sigfilesize)
+			goto err_cleanup;
+		insigdata = map_file(sigfile, sigfilesize);
+		if (insigdata == NULL)
+			DO_EXIT_1("Failed to open file: %s\n", infile);
+
+		if (sigfilesize >= sizeof(pemhdr) && !memcmp(insigdata, pemhdr, sizeof(pemhdr)-1))
+		{
+			sigbio = BIO_new_mem_buf(insigdata, sigfilesize);
+			sig = PEM_read_bio_PKCS7(sigbio, NULL, NULL, NULL);
+			BIO_free_all(sigbio);
+		}
+		else {
+			if (type == FILE_TYPE_PE) {
+				sig = extract_existing_pe_pkcs7(insigdata, peheader, pe32plus, 0, sigfilesize);
+			}
+			else if (type == FILE_TYPE_MSI) {
+#ifdef WITH_GSF
+				const unsigned char *p = (unsigned char*)insigdata;
+				sig = d2i_PKCS7(NULL, &p, sigfilesize);
+#else
+	                        DO_EXIT_1("libgsf is not available, msi support is disabled: %s\n", infile);
+#endif
+			}
+		}
+		if (!sig)
+			DO_EXIT_0("No valid signature found.");
+		goto add_only;
+	}
+
 	if (cmd != CMD_SIGN)
 		goto skip_signing;
 
@@ -2983,20 +3202,22 @@ int main(int argc, char **argv)
 	for(i = sk_X509_num(certs)-1; i>=0; i--)
 		PKCS7_add_certificate(sig, sk_X509_value(certs, i));
 
-	if (p7 == NULL) {
+	if (p7 == NULL || certs != p7->d.sign->cert) {
 		sk_X509_free(certs);
-	} else {
+		certs = NULL;
+	}
+	if (p7) {
 		PKCS7_free(p7);
 		p7 = NULL;
 	}
-	certs = NULL;
-	if (p7x == NULL) {
+	if (xcerts) {
 		sk_X509_free(xcerts);
-	} else {
+		xcerts = NULL;
+	}
+	if (p7x) {
 		PKCS7_free(p7x);
 		p7x = NULL;
 	}
-	xcerts = NULL;
 
 	get_indirect_data_blob(&p, &len, md, type, pagehash, indata, peheader, pe32plus, fileend);
 	len -= EVP_MD_size(md);
@@ -3026,14 +3247,20 @@ int main(int argc, char **argv)
 	ASN1_STRING_set(td7->d.other->value.sequence, buf, len+mdlen);
 	PKCS7_set_content(sig, td7);
 
+add_only:
+
 #ifdef ENABLE_CURL
 	/* add counter-signature/timestamp */
-	if (nturl && add_timestamp_authenticode(sig, turl, nturl, proxy))
+	if (nturl && add_timestamp_authenticode(sig, turl, nturl, proxy, noverifypeer))
 		DO_EXIT_0("authenticode timestamping failed\n");
-	if (ntsurl && add_timestamp_rfc3161(sig, tsurl, ntsurl, proxy, md))
+	if (ntsurl && add_timestamp_rfc3161(sig, tsurl, ntsurl, proxy, md, noverifypeer))
 		DO_EXIT_0("RFC 3161 timestamping failed\n");
 #endif
 
+	if (addBlob && add_unauthenticated_blob(sig))
+		DO_EXIT_0("Adding unauthenticated blob failed\n");
+
+
 #if 0
 	if (!PEM_write_PKCS7(stdout, sig))
 		DO_EXIT_0("PKCS7 output failed\n");
@@ -3076,7 +3303,7 @@ int main(int argc, char **argv)
 #ifdef WITH_GSF
 	} else if (type == FILE_TYPE_MSI) {
 		/* Only output signatures if we're signing. */
-		if (cmd == CMD_SIGN) {
+		if (cmd == CMD_SIGN || cmd == CMD_ADD || cmd == CMD_ATTACH) {
 			GsfOutput *child = gsf_outfile_new_child(outole, "\05DigitalSignature", FALSE);
 			if (!gsf_output_write(child, len, p))
 				DO_EXIT_1("Failed to write MSI 'DigitalSignature' signature to %s", infile);
@@ -3102,7 +3329,7 @@ int main(int argc, char **argv)
 skip_signing:
 
 	if (type == FILE_TYPE_PE) {
-		if (cmd == CMD_SIGN) {
+		if (cmd == CMD_SIGN || cmd == CMD_ADD || cmd == CMD_ATTACH) {
 			/* Update signature position and size */
 			(void)BIO_seek(outdata, peheader+152+pe32plus*16);
 			PUT_UINT32_LE(fileend, buf); /* Previous file end = signature table start */
@@ -3110,7 +3337,7 @@ skip_signing:
 			PUT_UINT32_LE(len+8+padlen, buf);
 			BIO_write(outdata, buf, 4);
 		}
-		if (cmd == CMD_SIGN || cmd == CMD_REMOVE)
+		if (cmd == CMD_SIGN || cmd == CMD_REMOVE || cmd == CMD_ADD || cmd == CMD_ATTACH)
 			recalc_pe_checksum(outdata, peheader);
 	} else if (type == FILE_TYPE_CAB) {
 		(void)BIO_seek(outdata, 0x30);
@@ -3121,8 +3348,49 @@ skip_signing:
 	BIO_free_all(hash);
 	hash = outdata = NULL;
 
-	printf(ret ? "Failed\n" : "Succeeded\n");
+	if (cmd == CMD_ATTACH) {
+		if (type == FILE_TYPE_PE) {
+			outdatasize = get_file_size(outfile);
+			if (!outdatasize)
+				DO_EXIT_0("Error verifying result.\n");
+			outdataverify = map_file(outfile, outdatasize);
+			if (!outdataverify)
+				DO_EXIT_0("Error verifying result.\n");
+			int sigpos = GET_UINT32_LE(outdataverify + peheader + 152 + pe32plus*16);
+			int siglen = GET_UINT32_LE(outdataverify + peheader + 152 + pe32plus*16 + 4);
+			ret = verify_pe_file(outdataverify, peheader, pe32plus, sigpos, siglen, leafhash);
+			if (ret) {
+				DO_EXIT_0("Signature mismatch.\n");
+			}
+		}
+		else if (type == FILE_TYPE_MSI)
+		{
+#ifdef WITH_GSF
+			GsfInput *src;
+			GsfInfile *ole;
 
+			src = gsf_input_stdio_new(outfile, NULL);
+			if (!src)
+				DO_EXIT_1("Error opening file %s", outfile);
+			ole = gsf_infile_msole_new(src, NULL);
+			g_object_unref(src);
+			ret = msi_verify_file(ole, leafhash);
+			g_object_unref(ole);
+			if (ret) {
+				DO_EXIT_0("Signature mismatch.\n");
+			}
+#else
+			DO_EXIT_1("libgsf is not available, msi support is disabled: %s\n", infile);
+#endif
+		}
+		else
+		{
+			DO_EXIT_1("Unknown input type for file: %s\n", infile);
+		}
+		printf("Signature successfully attached.\n");
+	}
+        else
+		printf(ret ? "Failed\n" : "Succeeded\n");
 	cleanup_lib_state();
 
 	return ret;
@@ -3133,18 +3401,16 @@ err_cleanup:
 		pass = NULL;
 	}
 	ERR_print_errors_fp(stderr);
+	if (certs && (p7 == NULL || certs != p7->d.sign->cert))
+		sk_X509_free(certs);
 	if (p7)
 		PKCS7_free(p7);
-	else if (certs)
-		sk_X509_free(certs);
 	if (p7x)
 		PKCS7_free(p7x);
-	else if (xcerts)
+	if (xcerts)
 		sk_X509_free(xcerts);
 	if (cert)
 		X509_free(cert);
-	if (certs)
-		sk_X509_free(certs);
 	if (pkey)
 		EVP_PKEY_free(pkey);
 	if (hash)
@@ -3162,4 +3428,6 @@ Local Variables:
    tab-width: 4
    indent-tabs-mode: t
 End:
+
+  vim: set ts=4 :
 */
